﻿using Proto;

namespace TestProject;

/// <summary>
/// 프로그램
/// </summary>
class Program
{
    //////////////////////////////////////////////////////////////////////////////////////////////////// Method
    ////////////////////////////////////////////////////////////////////////////////////////// Static
    //////////////////////////////////////////////////////////////////////////////// Private

    #region 프로그램 시작하기 - Main()

    /// <summary>
    /// 프로그램 시작하기
    /// </summary>
    private static void Main()
    {
        MessageHeader messageHeader = new MessageHeader
        (
            new Dictionary<string, string>
            {
                { "TraceID", Guid.NewGuid().ToString() },
                { "SpanID" , Guid.NewGuid().ToString() }
            }
        );

        ActorSystem actorSystem = new ActorSystem();

        RootContext rootContext = new RootContext
        (
            actorSystem,
            messageHeader,
            next => async (context, target, envelope) =>
            {
                MessageEnvelope newEnvelope = envelope
                    .WithHeader("TraceID"     , context.Headers.GetOrDefault("TraceID"))
                    .WithHeader("SpanID"      , Guid.NewGuid().ToString())
                    .WithHeader("ParentSpanID", context.Headers.GetOrDefault("SpanID"));

                Console.WriteLine("  1 루트 컨텍스트 발송자 미들웨어 진입");
                Console.WriteLine("  1 TraceID      : " + newEnvelope.Header.GetOrDefault("TraceID"     ));
                Console.WriteLine("  1 SpanID       : " + newEnvelope.Header.GetOrDefault("SpanID"      ));
                Console.WriteLine("  1 ParentSpanID : " + newEnvelope.Header.GetOrDefault("ParentSpanID"));

                await next(context, target, newEnvelope);

                Console.WriteLine("  1 루트 컨텍스트 발송자 미들웨어 이탈 - 송신은 비동기식입니다. 설계상 순서가 잘못되었습니다.");
            }
        );

        Props props = Props.FromFunc
        (
            context =>
            {
                if(context.Message is string)
                {
                    Console.WriteLine( "      3 액터 진입");
                    Console.WriteLine($"      3 TraceID      : {context.Headers.GetOrDefault("TraceID"     )}");
                    Console.WriteLine($"      3 SpanID       : {context.Headers.GetOrDefault("SpanID"      )}");
                    Console.WriteLine($"      3 ParentSpanID : {context.Headers.GetOrDefault("ParentSpanID")}");
                    Console.WriteLine($"      3 액터 수신    : {context.Message.GetType()}:{context.Message}");

                    context.Respond("세계!");

                    Console.WriteLine( "      3 액터 이탈");
                }

                return Task.CompletedTask;
            }
        )
        .WithReceiverMiddleware
        (
            receiver => async (context, envelope) =>
            {
                if(envelope.Message is string)
                {
                    MessageEnvelope newEnvelope = envelope
                        .WithHeader("TraceID"     , envelope.Header.GetOrDefault("TraceID"))
                        .WithHeader("SpanID"      , Guid.NewGuid().ToString())
                        .WithHeader("ParentSpanID", envelope.Header.GetOrDefault("SpanID"));

                    Console.WriteLine("    2 액터 수신자 미들웨어 진입");
                    Console.WriteLine("    2 TraceID      : " + newEnvelope.Header.GetOrDefault("TraceID"     ));
                    Console.WriteLine("    2 SpanID       : " + newEnvelope.Header.GetOrDefault("SpanID"      ));
                    Console.WriteLine("    2 ParentSpanID : " + newEnvelope.Header.GetOrDefault("ParentSpanID"));

                    await receiver(context, newEnvelope);

                    Console.WriteLine("    2 액터 수신자 미들웨어 이탈");
                }
                else
                {
                    await receiver(context, envelope);
                }
            }
        )
        .WithSenderMiddleware
        (
            sender => async (context, target, envelope) =>
            {
                MessageEnvelope newEnvelope = envelope
                    .WithHeader("TraceID"     , context.Headers.GetOrDefault("TraceID"))
                    .WithHeader("SpanID"      , Guid.NewGuid().ToString())
                    .WithHeader("ParentSpanID", context.Headers.GetOrDefault("SpanID"));

                Console.WriteLine("        4 액터 발송자 미들웨어 진입");
                Console.WriteLine("        4 TraceID      : " + newEnvelope.Header.GetOrDefault("TraceID"     ));
                Console.WriteLine("        4 SpanID       : " + newEnvelope.Header.GetOrDefault("SpanID"      ));
                Console.WriteLine("        4 ParentSpanID : " + newEnvelope.Header.GetOrDefault("ParentSpanID"));

                await sender(context, target, envelope);

                Console.WriteLine("        4 액터 발송자 미들웨어 이탈");
            }
        );

        PID pid = rootContext.Spawn(props);

        Task.Delay(500).Wait();

        Console.WriteLine("0 TraceID      : " + rootContext.Headers.GetOrDefault("TraceID"     ));
        Console.WriteLine("0 SpanID       : " + rootContext.Headers.GetOrDefault("SpanID"      ));
        Console.WriteLine("0 ParentSpanID : " + rootContext.Headers.GetOrDefault("ParentSpanID"));

        string result = rootContext.RequestAsync<string>(pid, "안녕하세요").Result;

        Console.WriteLine("결과 : " + result);

        Console.WriteLine("프로그램을 종료하기 위해서 아무 키나 눌러 주시기 바랍니다.");

        Console.ReadKey(false);
    }

    #endregion
}